;Countdown for school and university
; |-> display elapsed or remaining time of current lesson
; |-> display current lesson name
; '-> display DCF77 time and date
;
;AT90S2313 @ 4.194304 MHz crystal (Tiny2313 will work too)
;
; ,------------------,
; | NNNNPPP.P% WW-DD |
; | TTT # LLLLiHH:MM |
; '------------------'
;   N = Name  P = Percent   W = Weekday  D = Day (date)
;   T = Type  L = Location  H = Hour     M = Minute
;   # = Interval (7/14)     i = radio tower (while receiving DCF77)
;
;PORTS:
;  PB0-7 = LCD data bus
;  PD0   = DCF77 input
;  PD1   = switch: elapsed (1) / remaining (0)
;  PD5   = LCD 'RS'
;  PD6   = LCD 'E'
;
;VERSION HISTORY:
;  0.3   2006-10-29  Arne Rossius

.include "../../include/2313def.inc"

;LCD character constants
.equ	DCFoff = 0x00 ;user defined char
.equ	DCFon = 0x01 ;user defined char
.equ	Seven = 0x02 ;user defined char
.equ	Fourteen = 0x03 ;user defined char
.equ	ae = 0x04 ;user defined char
.equ	oe = 0x05 ;user defined char
.equ	ue = 0x06 ;user defined char
.equ	ae_ = 0xE1
.equ	sz = 0xE2
.equ	oe_ = 0xEF
.equ	ue_ = 0xF5

;day constants
.equ	MONDAY = 1
.equ	TUESDAY = 2
.equ	WEDNESDAY = 3
.equ	THURSDAY = 4
.equ	FRIDAY = 5
.equ	SATURDAY = 6
.equ	SUNDAY = 7
.equ	WEEKLY = 0x00
.equ	FORTN1 = 0x20
.equ	FORTN0 = 0x30

;Port constants
.equ	LCDData = PORTB
.equ	LCDControl = PORTD
.equ	RS = 5
.equ	E = 6
.equ	Inputs = PIND
.equ	DCFIn = 0
.equ	SwitchIn = 1

;registers
.def	lpmreg = R0
.def	dcf1 = R1
.def	dcf2 = R2
.def	dcf3 = R3
.def	dcf4 = R4
.def	dcf5 = R5
.def	second = R6
.def	minute = R7
.def	hour = R8
.def	day = R9
.def	wkday = R10
.def	op1l = R11
.def	op1h = R12
.def	op1e = R13
.def	op3l = R14
.def	op3h = R15

.def	temp = R16
.def	temp2 = R17
.def	w1 = R18
.def	op2e = R18 ;same as w1
.def	w2 = R19
.def	op3e = R19 ;same as w2
.def	dcfbits = R20
.def	dcftime = R21
.def	op2l = R22
.def	op2h = R23
.def	subsec = R24

.def	flags = R29
	.equ	DCFIn_LastState = DCFIn ;DCFIn = 0, see above
	.equ	DCFDisabled = 1
	.equ	StopDCF = 2
	.equ	TimeOK = 3
	.equ	DateOK = 4
.def	ZL = R30
.def	ZH = R31
.dseg

;===============================================================================

.cseg
.org 0x000
	rjmp	start
.org OVF0addr
	rjmp	t0ovf

lessons:
	;allowed time range is 00:00 to 18:00, shift with offset:
	.equ Offset = 6 ;time range 06:00 to 24:00 (24:00 will work, 00:00 or 24:01 will not!)

	;maximum length per lesson: 279 minutes (4:39)

;	    DAY|INTERVAL,	"Name",		"Typ",		"Room"
;		    TimeBegin,			TimeEnd
	.db MONDAY|FORTN1,	"KL  ",		ue,"2 ",	"Fo1 "
		.dw (18-Offset)*3600+00*60,	(19-Offset)*3600+30*60
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder

	.db TUESDAY|WEEKLY,	"GET1",		ue,"2 ",	"AM  "
		.dw (08-Offset)*3600+15*60,	(09-Offset)*3600+45*60
	.db TUESDAY|WEEKLY,	"HMA1",		"V2 ",		"Fo1 "
		.dw (10-Offset)*3600+00*60,	(11-Offset)*3600+30*60
	.db TUESDAY|WEEKLY,	"GGI1",		ue,"1 ",	"Ro  "
		.dw (12-Offset)*3600+45*60,	(13-Offset)*3600+30*60
	.db TUESDAY|FORTN0,	"GET1",		"K",ue,"2",	"RS5 "
		.dw (14-Offset)*3600+00*60,	(15-Offset)*3600+30*60
	.db TUESDAY|WEEKLY,	"GGI1",		"P2 ",		"CIP "
		.dw (16-Offset)*3600+15*60,	(18-Offset)*3600+30*60
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder

	.db WEDNESDAY|WEEKLY,	"PHY1",		"V2 ",		"AM  "
		.dw (08-Offset)*3600+00*60,	(09-Offset)*3600+30*60
	.db WEDNESDAY|WEEKLY,	"PHY1",		"Sem",		"Aul2"
		.dw (10-Offset)*3600+00*60,	(11-Offset)*3600+30*60
	.db WEDNESDAY|WEEKLY,	"GGI1",		"V2 ",		"Fo1 "
		.dw (14-Offset)*3600+00*60,	(15-Offset)*3600+30*60
	.db WEDNESDAY|WEEKLY,	"KL  ",		"V2 ",		"Fo1 "
		.dw (15-Offset)*3600+45*60,	(17-Offset)*3600+15*60
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder

	.db THURSDAY|WEEKLY,	"HMA1",		"V2 ",		"Fo1 "
		.dw (08-Offset)*3600+15*60,	(09-Offset)*3600+45*60
	.db THURSDAY|WEEKLY,	"HMA1",		ue,"2 ",	"Fo5 "
		.dw (10-Offset)*3600+00*60,	(11-Offset)*3600+30*60
	.db THURSDAY|WEEKLY,	"GET1",		"V2 ",		"Fo1 "
		.dw (11-Offset)*3600+45*60,	(13-Offset)*3600+15*60
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	
	.db FRIDAY|WEEKLY,	"PHY1",		ue,"2 ",	"Fo1 "
		.dw (08-Offset)*3600+00*60,	(09-Offset)*3600+30*60
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
	.dw 0,0,0,0,0,0,0,0 ;placeholder
lessons_end:

str_weekdays:	.db "MO", "TU", "WE", "TH", "FR", "SA", "SU"
str_nolesson:	.db "No lesson ", "right now "

;TODO:
;str_pause:	.db "    ", "Pause     " ;4 chars + 10 chars
;str_before:	.db "Nothing   ", "yet ...   " ;2x 10 chars
;str_after:	.db "Done for  ", "today :-) " ;2x 10 chars
;str_none:	.db "No work   ", "all day!  " ;2x 10 chars

lcdchars:
	.db 0x00, 0x00, 0x04, 0x0A, 0x04, 0x04, 0x04, 0x00 ;radio tower
	.db 0x0E, 0x11, 0x04, 0x0A, 0x04, 0x04, 0x04, 0x00 ;radio tower with wave
	.db 0x00, 0x0E, 0x02, 0x02, 0x02, 0x02, 0x00, 0x00 ;symbol "weekly"
	.db 0x00, 0x15, 0x15, 0x17, 0x11, 0x11, 0x00, 0x00 ;symbol "fortnightly"
	.db 0x0A, 0x00, 0x0E, 0x11, 0x1F, 0x11, 0x11, 0x00 ;A-umlaut
	.db 0x0A, 0x00, 0x0E, 0x11, 0x11, 0x11, 0x0E, 0x00 ;O-umlaut
	.db 0x0A, 0x00, 0x11, 0x11, 0x11, 0x11, 0x0E, 0x00 ;U-umlaut

;===============================================================================

start:
	;init stackpointer
	ldi	temp, LOW(RAMEND)
	out	SPL, temp

	;init ports
	ldi	temp, 0x7C
	out	DDRD, temp
	ldi	temp, 0x03
	out	PORTD, temp
	ldi	temp, 0xFF
	out	DDRB, temp

	;init timer 0
	ldi	temp, 0x05 ;clock source = CLK/1024
	out	TCCR0, temp
	ldi	temp, 1<<TOIE0 ;enable overflow int (16 Hz)
	out	TIMSK, temp

	;disable analog comparator
	ldi	temp, 1<<ACD
	out	ACSR, temp

	;init registers
; 	clr	year
; 	clr	month
	clr	day
	clr	wkday
	clr	hour
	clr	minute
	clr	second
	clr	subsec
	clr	dcfbits
	clr	dcftime
	clr	flags

	;DEBUG:
	;ldi temp, 0x30
	;mov day, temp
	;ldi temp, MONDAY
	;mov wkday, temp
	;ldi temp, 0x17
	;mov hour, temp
	;ldi temp, 0x59
	;mov minute, temp
	;ldi temp, 0x50
	;mov second, temp
	;ldi flags, 1<<DCFdisabled | 1<<TimeOK | 1<<DateOK

	;wait
	ldi	w1, 0 ;25.6 ms
	rcall	wait

	;init LCD
	rcall	lcdinit
	ldi ZL, LOW(str_nolesson*2)
	ldi ZH, HIGH(str_nolesson*2)
	ldi temp, 0x80
	ldi temp2, 10
	rcall str2lcd
	ldi temp, 0xC0
	ldi temp2, 10
	rcall str2lcd

	;enable interrupts
	sei

	;main loop: go to idle mode (sleep)
loop:
	in	temp, MCUCR
	ori	temp, 1<<SE
	out	MCUCR, temp
	sleep
	rjmp	loop

;===============================================================================

t0ovf:
	;timer 0 overflow interrupt (16 Hz)
	push	temp
	in	temp, SREG
	push	temp

	;DCF77 reception
	;don't evaluate DCF77 signal when reception is disabled
	sbrc	flags, DCFdisabled
	rjmp	dcf_end
	inc	dcftime
	;compare DCF77 input to last saved state
	in	temp, Inputs
	eor	temp, flags
	andi	temp, 1<<DCFIn
	brne	dcf_end ;inverted DCF77 signal (negative pulses)
	;breq	dcf_end ;non-inverted DCF77 signal (positive pulses)
	;take appropriate action for pos/neg edges
	sbrs	flags, DCFIn
	rjmp	dcf_posedge
	
	;negative edge ( 1 -> 0 )
	ldi	temp, 0xCA ;display "radio tower" symbol
	rcall	lcdcmd
	ldi	temp, DCFoff
	rcall	lcdchr
	andi	flags, ~(1<<DCFIn)
	sbrc	flags, StopDCF
	rjmp	dcf_end
	;increment bit number, exit if more than 59 bits were already received
	inc	dcfbits
	cpi	dcfbits, 60
	brsh	dcf_end
	rjmp	dcf_evalbit
	
dcf_posedge:
	;positive edge ( 0 -> 1 )
	ldi	temp, 0xCA ;display "radio tower with wave" symbol
	rcall	lcdcmd
	ldi	temp, DCFon
	rcall	lcdchr
	ori	flags, 1<<DCFIn
	;check if time was >1.2s (start of new telegram)
	cpi	dcftime, 19
	ldi	dcftime, 0
	brlo	dcf_end
	ldi	temp, 0xCA ;remove LCD symbol
	rcall	lcdcmd
	ldi	temp, ' '
	rcall	lcdchr
	cpi	dcfbits, 59
	ldi	dcfbits, 0
	brne	dcf_end
	rjmp	dcf_eval
dcf_end:



	;increment time (fallback quartz clock)
	;increment 1/16-seconds
	inc	subsec
	cpi	subsec, 16
	brsh	PC+2
	rjmp	t0ovf_end
	;increment seconds
	clr	subsec
	mov	temp, second
	rcall	bcdinc
	mov	second, temp
	cpi	temp, 0x60
	brlo	inctime_end
	;increment minutes
	clr	second
	mov	temp, minute
	rcall	bcdinc
	mov	minute, temp
	cpi	temp, 0x03 ;disable DCF77 reception if not successful after 3 minutes
	brne	inctime_min_end
	sbrc	flags, TimeOK
	ori	flags, 1<<DCFdisabled
	ldi	temp, 0xCA ;remove LCD symbol
	rcall	lcdcmd
	ldi	temp, ' '
	rcall	lcdchr
inctime_min_end:
	mov	temp, minute
	cpi	temp, 0x60
	brlo	inctime_end
	;increment hours
	clr	minute
	andi	flags, ~(1<<DCFdisabled) ;re-enable DCF77 reception
	mov	temp, hour
	rcall	bcdinc
	mov	hour, temp
	cpi	temp, 0x24
	brlo	inctime_end
	;new day: mark date as bad until the new one was received
	clr	hour
	andi	flags, ~(1<<DateOK)
	;TODO: write 0xFF (??) to day register
inctime_end:



	;display time and date
	sbrs	flags, DateOK
	rjmp	t0ovf_end ;nothing useful to display yet
	ldi	temp, 0xCB ;line 2, char 12
	rcall	lcdcmd
	mov	temp, hour
	rcall	bcd2lcd
	ldi	temp, ':'
	rcall	lcdchr
	mov	temp, minute
	rcall	bcd2lcd
	ldi	temp, 0x8B ;line 1, char 12
	rcall	lcdcmd
	mov	temp, wkday
	dec	temp
	lsl	temp, 2
	ldi	ZL, LOW(str_weekdays*2)
	ldi	ZH, HIGH(str_weekdays*2)
	add	ZL, temp
	ldi	temp, 0
	adc	ZH, temp
	lpm
	adiw	ZH:ZL, 1
	mov	temp, R0
	rcall	lcdchr
	lpm
	mov	temp, R0
	rcall	lcdchr
	ldi	temp, '-'
	rcall	lcdchr
	mov	temp, day
	rcall	bcd2lcd

	;calculate seconds since Offset:00:00
	mov	temp, second
	rcall	bcd2bin
	mov	op1l, temp
	clr	op1h
	mov	temp, hour
	rcall	bcd2bin
	subi	temp, Offset
;TODO	brcs	disp_before ;no lessons this early
	brcs	nolesson
	cpi	temp, 18
;TODO	brsh	disp_after ;no lessons this late
	brsh	nolesson
	clr	temp2
	ldi	op2l, LOW(3600)
	ldi	op2h, HIGH(3600)
	rcall	multiply ;op1 = op1 + (op2 * temp2:temp)
	mov	temp, minute
	rcall	bcd2bin
	clr	temp2
	ldi	op2l, LOW(60)
	ldi	op2h, HIGH(60)
	rcall	multiply ;op1 = op1 + (op2 * temp2:temp)
	mov	op2l, op1l
	mov	op2h, op1h



	;find current lesson (op2 = time)
	ldi	ZL, LOW(lessons*2)
	ldi	ZH, HIGH(lessons*2)
;TODO	andi	flags, ~(1<<LessonsStarted)
lesson_loop:
	lpm
	mov	temp, R0
	andi	temp, 0x0F
	cp	temp, wkday
	brne	lesson_next
	;TODO: check interval
	adiw	ZH:ZL, 12 ;check beginning
	lpm
	adiw	ZH:ZL, 1
	mov	op1l, R0
	lpm
	adiw	ZH:ZL, 1
	mov	op1h, R0
	cp	op2l, op1l
	cpc	op2h, op1h
	brlo	lesson_before ;time < begin
;FIXME - if needed, reset ZH:ZL first! - 	breq	lesson_start ;time = begin
;TODO	ori	flags, 1<<LessonsStarted
	lpm		;check end
	adiw	ZH:ZL, 1
	mov	op3l, R0
	lpm
	sbiw	ZH:ZL, 15
	mov	op3h, R0
	cp	op2l, op3l
	cpc	op2h, op3h
	brlo	lesson_found ;time < end
lesson_next:
	adiw	ZH:ZL, 16
	ldi	temp, HIGH(lessons_end*2)
	cpi	ZL, LOW(lessons_end*2)
	cpc	ZH, temp
	brlo	lesson_loop

lesson_before:
	;TODO - possibly a break, check flags[LessonsStarted]
nolesson:
	;no lesson found, clear lesson-specific display areas
	ldi	ZL, LOW(str_nolesson*2)
	ldi	ZH, HIGH(str_nolesson*2)
	ldi	temp, 0x80
	ldi	temp2, 10
	rcall	str2lcd
	ldi	temp, 0xC0
	ldi	temp2, 10
	rcall	str2lcd
	rjmp	t0ovf_end



lesson_start:
lesson_found:
	;lesson just started: display lesson information
	ldi	temp, 0xC3 ;interval (0x00 = weekly, 0x20/0x30 = fortnightly)
	rcall	lcdcmd
	ldi	temp, ' '
	rcall	lcdchr
	lpm
	adiw	ZH:ZL, 1
	ldi	temp, Seven
	sbrc	R0, 5 ;(matches 0x20 and 0x30)
	ldi	temp, Fourteen
	rcall	lcdchr
	ldi	temp, ' '
	rcall	lcdchr
	ldi	temp, 0x80 ;name
	ldi	temp2, 4
	rcall	str2lcd
	ldi	temp, 0xC0 ;type
	ldi	temp2, 3
	rcall	str2lcd
	ldi	temp, 0xC6 ;location
	ldi	temp2, 4
	rcall	str2lcd
;lesson_found:

	;calculate lesson length and elapsed time (op1 = start, op2 = time, op3 = end)
	sub	op2l, op1l ;op2 = elapsed = time - start
	sbc	op2h, op1h
	sub	op3l, op1l ;op3 = length = end - start
	sbc	op3h, op1h
	;calculate remaining time when switch is closed (input = 0)
	sbic	Inputs, SwitchIn
	rjmp	remaining_end
	mov	op1l, op3l ;op2 = remaining = length - elapsed
	mov	op1h, op3h
	sub	op1l, op2l
	sbc	op1h, op2h
	mov	op2l, op1l
	mov	op2h, op1h
remaining_end:

	;calculate percent elapsed/remaining
	ldi	temp, LOW(1000) ;multiply elaps/remain by 1000 (op2 = elaps/remain, op3 = length)
	ldi	temp2, HIGH(1000)
	clr	op1l
	clr	op1h
	clr	op1e
	rcall	multiply

	ldi	temp, 25 ;divide (elaps/remain*1000) by length (op1 = elaps/remain*1000, op3 = length)
	clr	op2l
	clr	op2h
	clr	op2e
	clr	op3e
	clc
divloop:
	rol	op1l
	rol	op1h
	rol	op1e
	dec	temp
	breq	divround
	rol	op2l
	rol	op2h
	rol	op2e
	sub	op2l, op3l
	sbc	op2h, op3h
	sbc	op2e, op3e
	brcc	divone
	add	op2l, op3l
	adc	op2h, op3h
	adc	op2e, op3e
	clc
	rjmp	divloop
divone:
	sec
	rjmp	divloop
divround:
	;TODO
divend:

	;display percent elapsed/remaining (op1 = per mille, so add a decimal point before the last digit)
	ldi	temp, 0x84
	rcall	lcdcmd
	ldi	op2l, LOW(1000) ;display 1000's
	ldi	op2h, HIGH(1000)
	rcall	bin2ascii
	cpi	temp, 0x30 ;remove leading zero
	brne	PC+2
	ldi	temp, ' '
	rcall	lcdchr
	mov	w1, temp
	ldi	op2l, 100 ;display 100's
	clr	op2h
	rcall	bin2ascii
	cpi	temp, 0x30 ;remove leading zero only if last digit also was removed
	brne	PC+4
	cpi	w1, ' '
	brne	PC+2
	ldi	temp, ' '
	rcall	lcdchr
	ldi	op2l, 10 ;display 10's
	rcall	bin2ascii
	rcall	lcdchr
	ldi	temp, '.' ;display decimal point
	rcall	lcdchr
	mov	temp, op1l ;display 1's
	ori	temp, 0x30
	rcall	lcdchr
	ldi	temp, '%' ;display percent sign
	rcall	lcdchr
	
t0ovf_end:
	;return to main program
	pop	temp
	out	SREG, temp
	pop	temp
	reti

;--------------------

;DCF77 reception (continued)

dcf_evalbit:
	;check bit length
	cpi	dcftime, 3 ;bit is 0: dcftime <= 2
	brlo	dcf_clearbit
	cpi	dcftime, 5 ;bit is 1: 3 <= dcftime <= 4
	brlo	dcf_setbit
	rjmp	dcf_end ;bit is erroneous: dcftime >= 5

	;save bit
dcf_setbit:
	sec ;set bit
	rjmp	dcf_savebit
dcf_clearbit:
	clc ;clear bit
dcf_savebit:
	ror	dcf5
	ror	dcf4
	ror	dcf3
	ror	dcf2
	ror	dcf1
	rjmp	dcf_end

;--------------------

;evaluate DCF77 telegram

dcf_eval:
	;check start bit
	sbrc	dcf1, 1
	rjmp	dcf_startbit_ok
	ori	flags, 1<<StopDCF
	rjmp	dcf_end
dcf_startbit_ok:

	;check date parity
	ldi	temp2, 0
	mov	temp, dcf3
	andi	temp, 0xFE
	rcall	parity
	mov	temp, dcf4
	rcall	parity
	mov	temp, dcf5
	andi	temp, 0x7F
	rcall	parity
	mov	temp, dcf5
	lsl	temp ;Bit7 -> C
	rol	temp ;C -> Bit0
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror
	
	;check hour parity
	ldi	temp2, 0
	mov	temp, dcf2
	andi	temp, 0xFC
	rcall	parity
	mov	temp, dcf3
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror
	
	;check minute parity
	ldi	temp2, 0
	mov	temp, dcf1
	andi	temp, 0xFC
	rcall	parity
	mov	temp, dcf2
	andi	temp, 0x01
	rcall	parity
	mov	temp, dcf2
	lsr	temp
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror

	;save minutes
	lsr	dcf2
	ror	dcf1
	lsr	dcf1
	mov	minute, dcf1
	
	;save hours
	lsr	dcf2
	mov	hour, dcf2
	
	;save day
	lsl	dcf3
	rol	dcf4 ;align month[5] weekday[3]
	rol	dcf5 ;align year
	lsr	dcf3 ;align day
	lsr	dcf3
	mov	day, dcf3
	
	;save weekday
	mov	wkday, dcf4
	ldi	temp, 0x07
	and	wkday, temp
	
; 	;save month
; 	lsr	dcf4
; 	lsr	dcf4
; 	lsr	dcf4
; 	mov	month, dcf4
; 	
; 	;save year
; 	mov	year, dcf5
	
	;set time & date ok, disable DCF77 reception
	clr	subsec
	clr	second
	ori	flags, 1<<TimeOK | 1<<DateOK | 1<<DCFdisabled
	ldi	temp, 0xCA ;remove LCD symbol
	rcall	lcdcmd
	ldi	temp, ' '
	rcall	lcdchr
	rjmp	dcf_end

dcf_parityerror:
	rjmp	dcf_end

;--------------------

parity:
	;calculate parity value of temp and add it to temp2
	lsr	temp
	brcc	PC+2
	inc	temp2
	cpi	temp, 0
	brne	parity
	andi	temp2, 0x01
	ret

;--------------------

bcdinc:
	;increment packed BCD value
	subi	temp, -7 ;increment and add 6
	brhc	PC+2 ;half-carry is inverted when using "subi rh, -k" instead of add!
	subi	temp, 6
	ret

;--------------------

bcd2bin:
	;convert packed BCD to binary
	mov	temp2, temp
	andi	temp, 0x0F
bcd2bin_loop:
	subi	temp2, 0x10
	brcs	bcd2bin_end
	subi	temp, -10
	rjmp	bcd2bin_loop
bcd2bin_end:
	ret

;--------------------

bin2ascii:
	;convert one digit of op1 to ascii (must be the most significiant digit!)
	ldi	temp, 0x30
bin2ascii_loop:
	cp	op1l, op2l
	cpc	op1h, op2h
	brlo	bin2ascii_end
	sub	op1l, op2l
	sbc	op1h, op2h
	inc	temp
	rjmp	bin2ascii_loop
bin2ascii_end:
	ret

;--------------------

multiply:
	;multiply op2h:op2l by temp2:temp and add result to op1e:op1h:op1l
	ldi	w1, 16
	clr	R0
mulloop:
	lsr	temp2
	ror	temp
	brcc	mulnoadd
	add	op1l, op2l
	adc	op1h, op2h
	adc	op1e, R0
mulnoadd:
	lsl	op2l
	rol	op2h
	rol	R0
	dec	w1
	brne	mulloop
	ret

;===============================================================================

str2lcd:
	;display a string from flash on the LCD (position = temp, length = temp2)
	rcall	lcdcmd
str2lcd_loop:
	lpm
	adiw	ZH:ZL, 1
	mov	temp, R0
	rcall	lcdchr
	dec	temp2
	brne	str2lcd_loop
	ret

;--------------------

bcd2lcd:
	;display a packed BCD value
	mov	temp2, temp
	swap	temp
	andi	temp, 0x0F
	ori	temp, 0x30
	rcall	lcdchr
	mov	temp, temp2
	andi	temp, 0x0F
	ori	temp, 0x30
	rcall	lcdchr
	ret

;--------------------

lcdinit:
	;initialize LCD
	ldi	temp, 0x30 ;8 bit mode
	rcall	lcdcmd
	ldi	w1, 50 ;5 ms
	rcall	wait
	rcall	lcdcmd
	rcall	lcdcmd
	ldi	temp, 0x3A ;8 bit mode, 2 lines, 5x7 font
	rcall	lcdcmd
	ldi	temp, 0x0C ;lcd power on, no cursor
	rcall	lcdcmd
	ldi	temp, 0x06 ;increment address pointer
	rcall	lcdcmd
	ldi	temp, 0x02 ;return home
	rcall	lcdcmd
	ldi	w1, 20 ;2 ms
	rcall	wait
	ldi	temp, 0x01 ;clear display
	rcall	lcdcmd
	ldi	w1, 20 ;2 ms
	rcall	wait

	;set user defined chars
	ldi	temp, 0x40 ;CGRAM, address 0x00
	rcall	lcdcmd
	ldi	ZL, LOW(lcdchars*2)
	ldi	ZH, HIGH(lcdchars*2)
	ldi	temp2, 64
lcdinit_loop:
	lpm
	adiw	ZH:ZL, 1
	mov	temp, R0
	rcall	lcdchr
	dec	temp2
	brne	lcdinit_loop
	ldi	temp, 0x80 ;DDRAM, address 0x00
	rcall	lcdcmd
	ret

;--------------------

lcdchr:
	;send a data byte to the LCD
	out	LCDData, temp
	sbi	LCDControl, RS
	sbi	LCDControl, E
	cbi	LCDControl, E
	ldi	w1, 1 ;100 us
	rcall	wait
	ret

;--------------------

lcdcmd:
	;send a command to the LCD
	out	LCDData, temp
	cbi	LCDControl, RS
	sbi	LCDControl, E
	cbi	LCDControl, E
	ldi	w1, 1 ;100 us
	rcall	wait
	ret

;--------------------

wait:
	;wait w1 x ~100 us
	ldi	w2, 0x85
wloop:
	dec	w2
	brne	wloop
	ldi	w2, 0x85
	dec	w1
	brne	wloop
	ret
	